""" Global application logging.

All modules use the same global logging object. No messages will be emitted
until the logger is started.

"""
from logging import Formatter
from logging import Logger as _Logger
from logging import NullHandler
from logging import StreamHandler


__all__ = "logger", "Logger"


class Logger(_Logger):
    """ Message logger.

    """
    LOGFMT = "%(asctime)s;%(levelname)s;%(name)s;%(message)s"

    def __init__(self, name=None):
        """ Initialize this logger.

        Loggers with the same name refer to the same underlying object. 
        Names are hierarchical, e.g. 'parent.child' defines a logger that is a 
        descendant of 'parent'.

        :param name: logger name (application name by default)
        """
        # With a NullHandler, client code may make logging calls without regard
        # to whether the logger has been started yet. The standard Logger API
        # may be used to add and remove additional handlers, but the
        # NullHandler should always be left in place. 
        super(Logger, self).__init__(name or __name__.split(".")[0])
        self.addHandler(NullHandler())  # default to no output
        return

    def start(self, level="WARN", stream=None):
        """ Start logging to a stream.

        Until the logger is started, no messages will be emitted. This applies
        to all loggers with the same name and any child loggers. 

        Multiple streams can be logged to by calling start() for each one.
        Calling start() more than once for the same stream will result in
        duplicate records to that stream.

        Messages less than the given priority level will be ignored. The
        default level conforms to the *nix convention that a successful run 
        should produce no diagnostic output. Call setLevel() to change the 
        logger's priority level after it has been stared. Available levels and 
        their suggested meanings:

            DEBUG - output useful for developers
            INFO - trace normal program flow, especially external interactions
            WARN - an abnormal condition was detected that might need attention
            ERROR - an error was detected but execution continued
            CRITICAL - an error was detected and execution was halted

        :param level: logger priority level
        :param stream: output stream (stderr by default)
        """
        handler = StreamHandler(stream)
        handler.setFormatter(Formatter(self.LOGFMT))
        self.addHandler(handler)
        self.setLevel(level.upper())
        return

    def stop(self):
        """ Stop logging with this logger.

        """
        for handler in self.handlers[1:]:
            # Remove everything but the NullHandler.
            self.removeHandler(handler)
        return


logger = Logger()



""" Implement the readwav command.

"""
import csv

from uuid import uuid4
from pathlib import Path
from guano import GuanoFile
from ..core.logger import logger


def main(wav_dir, metadata_file) -> str:
    """ Execute the command.

    :param wav_dir: directory containing one more more wav files
    :param metadata_file: full path to GUANO metadata output file
    """
    logger.debug("executing readwav command")

    guano_strict = {'GUANO|Version': '',
                    'Filter HP': '',
                    'Filter LP': '',
                    'Firmware Version': '',
                    'Hardware Version': '',
                    'Humidity': '',
                    'Length': '',
                    'Loc Accuracy': '',
                    'Loc Elevation': '',
                    'Loc Position': '',
                    'Make': '',
                    'Model': '',
                    'Note': '',
                    'Original Filename': '',
                    'Samplerate': '',
                    'Serial': '',
                    'Species Auto ID': '',
                    'Species Manual ID': '',
                    'Tags': '',
                    'TE': '',
                    'Temperature Ext': '',
                    'Temperature Int': '',
                    'Timestamp': ''}

    combined_metadata = {}
    combined_metadata.update(guano_strict)

    with open(metadata_file, 'w', newline='') as output_file:
        for file in Path(wav_dir[0]).glob('*.[Ww][Aa][Vv]'):
            try:
                gf = GuanoFile(Path(wav_dir[0]).joinpath(str(file.name)))
            except ValueError:
                logger.warn(file.name + ' is not GUANO compliant')
            else:
                guano_metadata = {key: value for key, value in gf.items()}
                combined_metadata.update(guano_metadata)
                combined_metadata.update({'ABCD|uuid': uuid4()})
                combined_metadata.update({'Original Filename': file.name})

                if output_file.tell() == 0:
                    writer = csv.DictWriter(output_file, dialect=csv.excel, fieldnames=combined_metadata.keys())
                    writer.writeheader()

                writer.writerow(combined_metadata)

    return 'Output file written to: ' + output_file.name



""" Implementation of the command line interface.

"""
import os
from argparse import ArgumentParser
from inspect import getfullargspec

from . import __version__
from .api import hello
from .api import readwav
from .core.config import config
from .core.logger import logger

__all__ = "main",


def main(argv=None) -> int:
    """ Execute the application CLI.

    :param argv: argument list to parse (sys.argv by default)
    :return: exit status
    """
    args = _args(argv)
    logger.start(args.warn or "DEBUG")  # can't use default from config yet
    logger.debug("starting execution")
    config.load(args.config)
    config.core.config = args.config
    if args.warn:
        config.core.logging = args.warn
    logger.stop()  # clear handlers to prevent duplicate records
    logger.start(config.core.logging)
    command = args.command
    args = vars(args)
    spec = getfullargspec(command)
    if not spec.varkw:
        # No kwargs, remove unexpected arguments.
        args = {key: args[key] for key in args if key in spec.args}
    try:
        command(**args)
    except RuntimeError as err:
        logger.critical(err)
        return 1
    logger.debug("successful completion")
    return 0


def _args(argv):
    """ Parse command line arguments.

    :param argv: argument list to parse
    """
    parser = ArgumentParser()
    parser.add_argument("-c", "--config", action="append",
                        help="config file [etc/config.yml]")
    parser.add_argument("-v", "--version", action="version",
                        version="guanoctl {:s}".format(__version__),
                        help="print version and exit")
    parser.add_argument("-w", "--warn", default="WARN",
                        help="logger warning level [WARN]")
    common = ArgumentParser(add_help=False)  # common subcommand arguments
    # common.add_argument('wav_dir')
    # common.add_argument("--name", "-n", default="World", help="greeting name")
    subparsers = parser.add_subparsers(title="subcommands")
    _hello(subparsers, common)
    _readwav(subparsers, common)
    args = parser.parse_args(argv)
    if not args.config:
        # Don't specify this as an argument default or else it will always be
        # included in the list.
        args.config = "etc/config.yml"
    return args


def _hello(subparsers, common):
    """ CLI adaptor for the api.hello command.

    :param subparsers: subcommand parsers
    :param common: parser for common subcommand arguments
    """
    parser = subparsers.add_parser("hello", parents=[common])
    parser.set_defaults(command=hello)
    return


def _readwav(subparsers, common):
    """CLI adapter for the api.readwav command.

    :param subparsers: subcommand parsers
    :param common: parser for common subcommand arguments
    """
    parser = subparsers.add_parser('readwav', parents=[common])
    parser.add_argument('wav_dir', nargs=1, help='Full path of a directory containing one or more sound files.')
    parser.add_argument('-o', '--metadata_file', type=str, default=os.path.join(os.getcwd(), 'guano_metadata.csv'))
    parser.set_defaults(command=readwav)
    return


# Make the module executable.

if __name__ == "__main__":
    try:
        status = main()
    except Exception:  # 'Exception' added by DS
        logger.critical("shutting down due to fatal error")
        raise  # print stack trace
    else:
        raise SystemExit(status)



""" Main application entry point.

    python -m guanoctl  ...

"""
def main():
    """ Execute the application.

    """
    raise NotImplementedError


# Make the script executable.

if __name__ == "__main__":
    raise SystemExit(main())



""" Implement the hello command.

"""
from ..core.logger import logger


def main(name="World") -> str:
    """ Execute the command.
    
    :param name: name to use in greeting
    """
    logger.debug("executing hello command")
    return "Hello, {:s}!".format(name)  # TODO: use f-string for Python 3.6+



""" Setup script for the guanoctl application.

"""
from os import walk
from pathlib import Path

from setuptools import find_packages
from setuptools import setup


_config = {
    "name": "guanoctl",
    "long_description": "A command-line tool for guano-py (https://github.com/riggsd/guano-py)",
    "url": "",
    "author": "Danny Squire",
    "author_email": "danny@el-fuego.net",
    "package_dir": {"": "src"},
    "packages": find_packages("src"),
    "entry_points": {
        "console_scripts": ("guanoctl = guanoctl.cli:main",),
    },
    "data_files": ("etc/",),
}


def main() -> int:
    """ Execute the setup command.

    """
    def data_files(*paths):
        """ Expand path contents for the `data_files` config variable.  """
        for path in map(Path, paths):
            if path.is_dir():
                for root, _, files in walk(str(path)):
                    yield root, tuple(str(Path(root, name)) for name in files)
            else:
                yield str(path.parent), (str(path),)
        return

    def version():
        """ Get the local package version. """
        namespace = {}
        path = Path("src", _config["name"], "__version__.py")
        exec(path.read_text(), namespace)
        return namespace["__version__"]

    _config.update({
        "data_files": list(data_files(*_config["data_files"])),
        "version": version(),
    })
    setup(**_config)
    return 0


# Make the script executable.

if __name__ == "__main__":
    raise SystemExit(main())



""" Global application configuration.

This module defines a global configuration object. Other modules should use
this object to store application-wide configuration values.

"""
from re import compile
from yaml import safe_load

from .logger import logger


__all__ = "config", "YamlConfig"


class _AttrDict(dict):
    """ A dict-like object with attribute access.

    """
    def __getitem__(self, key):
        """ Access dict values by key.

        :param key: key to retrieve
        """
        value = super(_AttrDict, self).__getitem__(key)
        if isinstance(value, dict):
            # For mixed recursive assignment (e.g. `a["b"].c = value` to work
            # as expected, all dict-like values must themselves be _AttrDicts.
            # The "right way" to do this would be to convert to an _AttrDict on
            # assignment, but that requires overriding both __setitem__
            # (straightforward) and __init__ (good luck). An explicit type
            # check is used here instead of EAFP because exceptions would be
            # frequent for hierarchical data with lots of nested dicts.
            self[key] = value = _AttrDict(value)
        return value

    def __getattr__(self, key):
        """ Get dict values as attributes.

        :param key: key to retrieve
        """
        return self[key]

    def __setattr__(self, key, value):
        """ Set dict values as attributes.

        :param key: key to set
        :param value: new value for key
        """
        self[key] = value
        return


class YamlConfig(_AttrDict):
    """ Store YAML configuration data.

    Data can be accessed as dict values or object attributes.

    """
    def __init__(self, path=None, root=None, macros=None):
        """ Initialize this object.

        :param path: config file path to load
        :param root: place config values at this root
        :param macros: macro substitutions
        """
        super(YamlConfig, self).__init__()
        if path:
            self.load(path, root, macros)
        return

    def load(self, path, root=None, macros=None):
        """ Load data from YAML configuration files.

        Configuration values are read from a sequence of one or more YAML
        files. Files are read in the given order, and a duplicate value will
        overwrite the existing value. If 'root' is specified the config data
        will be loaded under that attribute instead of the dict root.

        The optional 'macros' argument is a dict-like object to use for macro
        substitution in the config files. Any text matching "%key;" will be
        replaced with the value for 'key' in 'macros'.

        :param path: config file path to load
        :param root: place config values at this root
        :param macros: macro substitutions
        """
        def replace(match):
            """ Callback for re.sub to do macro replacement. """
            # This allows for multi-pattern substitution in a single pass.
            return macros[match.group(0)]

        macros = {r"%{:s};".format(key): val for (key, val) in
                  macros.items()} if macros else {}
        regex = compile("|".join(macros) or r"^(?!)")
        for path in [path] if isinstance(path, str) else path:
            with open(path, "r") as stream:
                # Global text substitution is used for macro replacement. Two
                # drawbacks of this are 1) the entire config file has to be
                # read into memory first; 2) it might be nice if comments were
                # excluded from replacement. A more elegant (but complex)
                # approach would be to use PyYAML's various hooks to do the
                # substitution as the file is parsed.
                logger.info("reading config data from '{:s}'".format(path))
                yaml = regex.sub(replace, stream.read())
            data = safe_load(yaml)
            try:
                if root:
                    self.setdefault(root, {}).update(data)
                else:
                    self.update(data)
            except TypeError:  # data is None
                logger.warning("config file {:s} is empty".format(path))
        return


config = YamlConfig()
